(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`Hamiltonian`",{
  (*"PhysicalConstants`","Units`",*)
  "AtomicDensityMatrix`",
  "AtomicDensityMatrix`Common`",
  "AtomicDensityMatrix`CouplingCoefficients`",
  "AtomicDensityMatrix`TensorAlgebra`",
  "AtomicDensityMatrix`AtomicSystem`",
  "AtomicDensityMatrix`ExperimentalSystem`",
  "AtomicDensityMatrix`DensityMatrix`",
  "AtomicDensityMatrix`OpticalField`",
  "AtomicDensityMatrix`RotatingWaveApproximation`",
  "AtomicDensityMatrix`WignerEckart`", 
  "AtomicDensityMatrix`WignerRotate`"}]

Begin["`Private`"]

(*Hamiltonians*)

(*Unperturbed Hamiltonian*)

InternalHamiltonian[system:{_AtomicState..}] := 
	WignerEckart[system,{Energy,0,0}]

Energies[system:{_AtomicState..},crit_] := 
	Union[WignerEckart[#,{Energy,0,0},#]&/@SelectStates[system,crit]]

TransitionFrequencies[system:{_AtomicState..},label1_,label2_] :=
	Transpose@Expand@Outer[Plus,Energies[system,label2],-Energies[system,label1]]

(*E1 Hamiltonian*)

(*Subscript[H, l]=-d\[CenterDot]E*)

E1Hamiltonian[system:{_AtomicState..} /; AtomicSystemType[system] === "Toy", 
	efield_ /; CartesianQ[efield]||ScalarQ[efield]] :=  
	Module[{components = DeleteCases[EnsureList[efield],_?PossibleZeroQ]},
		If[
			components === {},
			WignerEckart[system,{Zero,0,0}],
			-ComplexExpand[Re[First@components]]WignerEckart[system,{Dipole,1,0}]
		]
	]

E1Hamiltonian[system:{_AtomicState..}, efield_?CartesianQ] := 
	-ComplexExpand[Re[efield]]\[CenterDot]ToCartesian[WignerEckart[system,{Dipole,1}]]

(*E1Hamiltonian[system:{_AtomicState..},{efield_,label_},opts___Rule]:=E1Hamiltonian[system,efield,Label->label,opts]*)

(*Magnetic field Hamiltonian*)

(*Subscript[H, B]=-\[Mu]\[CenterDot]B*)

MagneticHamiltonian[system:{_AtomicState..},bf_?CartesianQ] := 
	-bf\[CenterDot]ToCartesian[WignerEckart[system,{MagneticMoment,1}]]

MagneticHamiltonian[system:{_AtomicState..},bf_?ScalarQ]:=MagneticHamiltonian[system,{0,0,bf}]


(*Effective electric field Hamiltonian*)


(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[*)(*Subscript[H, Stark]={*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"]\[CircleTimes]*)(*SuperscriptBox[*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"],"*"]}, \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

PolarizabilityHamiltonian[system:{_AtomicState..},efield_?CartesianQ,k_]:=
	SphericalTensorProduct[ToCovariant@efield,ToCovariant@EC@efield,{k}]\[CenterDot]ECT@WignerEckart[system,{Polarizability,k}]

(*PolarizabilityHamiltonian[system:{_AtomicState..},{efield_?CartesianQ,label_},k_,opts___Rule]:=
PolarizabilityHamiltonian[system,efield,k,Label->label,opts]*)

PolarizabilityHamiltonian[system:{_AtomicState..},efield_?ScalarQ,k_] := 
	PolarizabilityHamiltonian[system,{0,0,efield},k]

PolarizabilityHamiltonian[system:{_AtomicState..},efield_] := 
	Sum[PolarizabilityHamiltonian[system,efield,k],{k,0,2}]


(*Transition Polarizability Hamiltonian*)

TransitionPolarizabilityHamiltonian[system:{_AtomicState..},opticalfield_?CartesianQ,staticfield_?CartesianQ,k_]:=
	SphericalTensorProduct[
		ToCovariant@staticfield,
		ToCovariant@ComplexExpand@Re@opticalfield,{k}
	]\[CenterDot]ECT@WignerEckart[system,{TransitionPolarizability,k}]

TransitionPolarizabilityHamiltonian[system:{_AtomicState..},opticalfield_, staticfield_] := 
	Sum[TransitionPolarizabilityHamiltonian[system,opticalfield, staticfield,k],{k,0,2}]


(*Total Hamiltonian*)


Options[Hamiltonian]={ElectricField->0,MagneticField->0,Interaction->Automatic}

Hamiltonian[system:{_AtomicState..},expSystem:{_Region..},opts:OptionsPattern[]] := 
	Hamiltonian[system,#,opts]&/@expSystem

Hamiltonian[system:{_AtomicState..},region_Region,opts:OptionsPattern[]] := 
	Hamiltonian[system,
		ElectricField->If[DefinedQ[region,ElectricField],ElectricField[region], {0,0,0}]
			+If[DefinedQ[region,OpticalParameters],OpticalField@@(OpticalParameters[region]), {0,0,0}],
		MagneticField->If[DefinedQ[region,MagneticField],MagneticField[region], {0,0,0}], 
		opts
	]

Hamiltonian[system:{_AtomicState..},OptionsPattern[]] :=
    Module[{EF=OptionValue[ElectricField],BF=OptionValue[MagneticField],
    	    type=OptionValue[Interaction],zero = WignerEckart[system,{Zero,0,0}]},
        If[ type===All,
            type = {"Internal","ElectricDipole","MagneticDipole","Polarizability"(*,"TransitionPolarizability"*)}
        ];
        If[ type===Automatic,
            type = {"Internal","ElectricDipole","MagneticDipole"}
        ];
        type = EnsureList[type];
        If[ EF===0, EF = {0,0,0}];
        EF = If[ MatchQ[EF,{{_?CartesianQ,_}..}], EF, {EF}];
        If[ MemberQ[type,"Internal"|Internal],
            InternalHamiltonian[system],
            zero
        ] + 
        If[ MemberQ[type,"ElectricDipole"|E1],
            Total[(E1Hamiltonian[system,#1]&)/@EF],
            zero
        ] + 
        If[ MemberQ[type,"MagneticDipole"|Magnetic]&&AtomicSystemType[system] =!= "Toy",
            MagneticHamiltonian[system,BF],
            zero
        ] + 
        If[ MemberQ[type,"Polarizability"|Polarizability],
            Total[(PolarizabilityHamiltonian[system,#1]&)/@EF],
            zero
        ]
        (* + 
        If[ MemberQ[type,"TransitionPolarizability"|TransitionPolarizability],
            TransitionPolarizabilityHamiltonian[system,Cases[EF,{x_,AC}->x,1][[1]],Cases[EF,{x_,DC}->x,1][[1]]],
            zero
        ]*)
    ]

End[]

EndPackage[]
