(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`LiouvilleEquation`",{
  (*"PhysicalConstants`","Units`",*)
  "AtomicDensityMatrix`",
  "AtomicDensityMatrix`Common`",
  "AtomicDensityMatrix`CouplingCoefficients`",
  "AtomicDensityMatrix`TensorAlgebra`",
  "AtomicDensityMatrix`AtomicSystem`",
  "AtomicDensityMatrix`ExperimentalSystem`",
  "AtomicDensityMatrix`DensityMatrix`",
  "AtomicDensityMatrix`WignerEckart`", 
  "AtomicDensityMatrix`WignerRotate`"}]

Begin["`Private`"]

(* ::Subsection:: *)
(*Relaxation and repopulation*)


(*Relaxation matrix*)

IntrinsicRelaxation[system:{_AtomicState..}]=WignerEckart[system,{NaturalWidth,0,0}]

TransitRelaxation[system:{_AtomicState..},gt_]:=gt WignerEckart[system,{Identity,0,0}]

(*Transit ground-state repopulation matrix*)

TransitRepopulation[system:{_AtomicState..},gt_]:=Module[{g=SublevelMultiplicity[GroundStates[system]]},
	gt WignerEckart[system,{GroundState,0,0}]/If[g==0,1,g]
]



(* ::Subsubsection:: *)
(*Repopulation due to spontaneous decay*)


Options[OpticalRepopulation]=Join[{HyperfineCoherenceTransfer->False,ExpandDipoleRME->True,Method->Automatic},Options[DMVariables]]

OpticalRepopulation[system:{_AtomicState..},opts:OptionsPattern[]]:=
	Module[{newopts},
		newopts=Join[DeleteCases[Join[{opts},Options[OpticalRepopulation]],_->Automatic],Options[DensityMatrix]];
		Switch[Representation/.newopts,
			"Zeeman",If[OptionValue[Method]==="Old", OldOpticalRepopulationZ, OpticalRepopulationZ][system,Sequence@@newopts],
			PolarizationMoments,OpticalRepopulationPM[system,Sequence@@newopts]
		]
	]



OpticalRepopulationZ[system:{_AtomicState..},opts:OptionsPattern[OpticalRepopulation]]:=
	Module[{d,branchingMat,degeneracyBoole,dm,dmopts,branchingBoole},
		dmopts=FilterRules[{opts},Options[DensityMatrix]];
		(*cartesian components of the electric dipole operator*)	
		d = ToCartesian@WignerEckart[system, {Dipole, 1}] /. ReducedME[_, {Dipole, 1}, _] -> 1;
 		(* Matrix of branching rates *)
 		branchingMat = (2 J[system] + 1) NaturalWidth[system] DirectSelfProduct[BranchingRatio[#2][#1] &, system];
 		branchingBoole = Map[Boole[# =!= 0] &, branchingMat, {2}];
 		(* Only allow transfer of coherences from a pair of degenerate levels to another pair of degenerate levels *)
 		degeneracyBoole = DirectSelfProduct[
 			Composition[Boole, PossibleZeroQ, Subtract], 
 			Energy@system + If[OptionValue[HyperfineCoherenceTransfer],0,HyperfineShift@system]
 		];
 		dm = DensityMatrix[system,Sequence@@dmopts] degeneracyBoole;
		Expand[degeneracyBoole MatrixMultiply[(# Transpose[branchingMat]& /@ d).dm, # branchingBoole& /@ d]]
	]	
	
	
OldOpticalRepopulationZ[system:{_AtomicState..},opts___Rule]:=Module[{dc,br,w,w1,dm,len,hct,dmopts},
	dmopts=FilterRules[{opts},Options[DensityMatrix]];
	(*dc is the cartesian components of the electric dipole operator, transposed to a matrix of vectors *)	
	dc=Transpose[ToCartesian@WignerEckart[system,{Dipole,1}],{3,1,2}];
	dc=If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,dc],dc];
	br=DirectSelfProduct[Boole[BranchingRatio[#1][#2]=!=0]&,system];
	hct=HyperfineCoherenceTransfer/.{opts}/.Options[OpticalRepopulation];
	w=DirectSelfProduct[Subtract,Energy/@system];
	w1=If[hct,w,DirectSelfProduct[Subtract,Tr[WignerEckart[system,{Energy,0,0}],List]]];
	dm=DensityMatrix[system,Sequence@@dmopts];
	len=Length[system];
	Table[
		If[PossibleZeroQ[w1[[m,n]]],
			Sum[
				If[PossibleZeroQ[w1[[p,q]]]&&br[[m, q]]!=0 ,
					4/3 br[[m,q]] w[[q,m]]^3 (dc[[p,n]].dc[[m,q]])dm[[q,p]],
					0
				],
				{p,len},{q,len}
			],
			0
		],
	{m,len},{n,len}]
]

OpticalRepopulationPM[system:{_AtomicState..},opts___Rule] := 
	Module[
		{levels=RemoveZeemanSublevels[system]},
		upperlevels= Select[levels,!GroundStateQ[#]&];
		pm=PolarizationMoments[upperlevels];
		If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation], ExpandDipoleRME[system,#],#]&@Table[
			4/3 Sum[If[Abs[F[le1]-F[le2]]<= k<= F[le1]+F[le2],(Energy[le1]-Energy[l1])^3 ReducedME[l1,{Dipole,1},le1]Component[upperlevels,pm,{le1,le2},k]ReducedME[le2,{Dipole,1},l2] (-1)^(F[le1]+F[le2]+k+1) SJS[{k,F[l2],F[l1]},{1,F[le1],F[le2]}],
			Transpose[{ConstantArray[0,2k+1]}]],{le1,upperlevels},{le2,upperlevels}],
		{l1,levels},{l2,levels},{k,Abs[F[l1]-F[l2]],F[l1]+F[l2]}]
	]

Options[OpticalRepopulationOperator]={HyperfineCoherenceTransfer->False,ExpandDipoleRME->True}

OpticalRepopulationOperator[system:{_AtomicState..},opts___Rule]:=
	Module[{dm,rep,or},
		dm=DensityMatrix[system];
		rep=Thread[Flatten[dm]->0];
		or=OpticalRepopulation[system,opts];
		Transpose[Map[(or /. #->1 /. rep)&,dm,{2}],{3,4,2,1}]
	]

(*The OpticalRepopulation matrix is equal to Map[Tr[#\[CenterDot]DensityMatrix[System]]&,OpticalRepopulationOperator[System],{2}]*)

(*ExperimentalRegion input*)

Relaxation[as:{_AtomicState..},es:{_Region..}]:=
	(
	IntrinsicRelaxation[as]
	+TransitRelaxation[as,TransitRate[VoidRegion][#]]
	+TransitRelaxation[as,Sum[TransitRate[l][#],{l,es}]]
	)&/@es

Repopulation[as:{_AtomicState..},es:{_Region..}]:=
	(
	TransitRepopulation[as,Density[#]TransitRate[VoidRegion][#]]
	+OpticalRepopulation[as,DMLabel->RegionLabel[#]]
	+Sum[Density[#]/Density[l] TransitRate[l][#]DensityMatrix[as,DMLabel->RegionLabel[l]],{l,es}]
	)&/@es

(*FluorescenceOperator*)

Options[FluorescenceOperator]={ExpandDipoleRME->True}

FluorescenceOperator[system:{_AtomicState..},freq_,th_,phi_,opts___Rule]:=
Module[{dc,gr,ikk},
	dc=ToCartesian@WignerEckart[system,{Dipole,1}];
	dc=If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,dc],dc];
	gr=DiagonalMatrix[Boole/@GroundStateQ/@system];
	(*ikk=IdentityMatrix[3]-DirectSelfProduct[CoordinatesToCartesian[{1,th,phi},Spherical]];*)
	ikk=IdentityMatrix[3]-DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
	freq^4/(2\[Pi]) Sum[dc[[i]].gr.dc[[j]] ikk[[i,j]],{i,3},{j,3}]
]

FluorescenceOperator[system:{_AtomicState..},freq_,opts___Rule]:=
Module[{dc,gr},
	dc=ToCartesian@WignerEckart[system,{Dipole,1}];
	dc=If[ExpandDipoleRME/.{opts}/.Options[OpticalRepopulation],ExpandDipoleRME[system,dc],dc];
	gr=DiagonalMatrix[Boole/@GroundStateQ/@system];
	4/3 freq^4 Sum[dc[[i]].gr.dc[[i]],{i,3}]
]


(*Liouville equation*)

(*Make the Liouville equation*)

LiouvilleRHS[system:{_AtomicState..},H_?OperatorQ,GMatrix_?OperatorQ,LMatrix_?OperatorQ,opts___Rule]:=
-I Commutator[H,DensityMatrix[system,opts]]-Anticommutator[DensityMatrix[system,opts],GMatrix]/2+LMatrix

Options[LiouvilleEquation]=Options[DMVariables]

LiouvilleEquation[system:{_AtomicState..},H_?OperatorQ,GMatrix_?OperatorQ,LMatrix_?OperatorQ,opts___Rule]:=
	Module[{rep,lhs,rhs,newopts,tv,cev,H1,GMatrix1,LMatrix1},
		newopts = DeleteCases[Join[{opts},Options[LiouvilleEquation]],_->Automatic];
		{rep,tv,cev} = {Representation,TimeVariable,ComplexExpandVariables}/.newopts/.Options[DensityMatrix];
		lhs = D[DensityMatrix[system,Sequence@@newopts],tv];
		{H1,GMatrix1,LMatrix1} = {H,GMatrix,LMatrix}/.ExpandDMVariablesRule[system,Sequence@@newopts];
		rhs = LiouvilleRHS[system,H1,GMatrix1,LMatrix1,Sequence@@newopts];
		If[rep===PolarizationMoments,
		MapThread[Equal,{Simplify@Flatten@Decomposition[system,lhs],Flatten@Decomposition[system,rhs]},1],
			If[cev||cev===Subscript,
				DeleteCases[Flatten@ERI@UpperTriangle[#],True]&,
				Flatten
			]@MapThread[Equal,{lhs,rhs},2]
		]
	]

LiouvilleEquation[system:{_AtomicState..},H_?OperatorQ,GMatrix_?OperatorQ,opts___Rule]:=LiouvilleEquation[system,H,GMatrix,0IdentityMatrix[Length[system]],opts]

LiouvilleEquation[system:{_AtomicState..},H_?OperatorQ,opts___Rule]:=LiouvilleEquation[system,H,0IdentityMatrix[Length[system]],0IdentityMatrix[Length[system]],opts]

LiouvilleEquation[
	as:{_AtomicState..},
	es:{_Region..},H:{_?OperatorQ..},
	GMatrix:{_?OperatorQ..},
	LMatrix:{_?OperatorQ..},
	opts:OptionsPattern[]]:=
	Flatten[MapThread[LiouvilleEquation[as,#2,#3,#4,DMLabel->RegionLabel[#1],opts]&,{es,H,GMatrix,LMatrix}],1]

(*ExpandedLiouvilleEquation[system:{_AtomicState..},args__?OperatorQ,opts___Rule]:=SeparateDensityMatrix[system,UpperTriangle@LiouvilleEquation[system,args,opts]]*)

LiouvilleEquation[sys:{_AtomicState..},H_?ContravariantDecompositionMatrixQ,
	G_?ContravariantDecompositionMatrixQ,L_?ContravariantDecompositionMatrixQ,opts___Rule] :=
    Module[ {newopts = Join[DeleteCases[Join[{opts},Options[LiouvilleEquation]],_->Automatic],Options[DensityMatrix]],levels = RemoveZeemanSublevels[sys],pm,n,f,component},
        pm = PolarizationMoments[levels,opts];
        n = Length[levels];
        f = F/@levels;
        component[op_,{i1_,i2_},k_,q_] :=
            op[[i1,i2,k-Abs[f[[i1]]-f[[i2]]]+1,-q+k+1,1]];
        Flatten@Table[
        Transpose[{Table[
        D[component[pm,{i1,i2},k,q],TimeVariable/.newopts]==
        Sum[-I (-1)^(f[[i1]]+f[[i2]]+k) Sqrt[(2k1+1)(2k2+1)]CG[{k1,q1},{k2,q2},{k,q}]SJS[{k1,k2,k},{f[[i2]],f[[i1]],f[[i3]]}]
        *(
        (component[H,{i1,i3},k1,q1]-I/2 component[G,{i1,i3},k1,q1])component[pm,{i3,i2},k2,q2]
        -component[pm,{i1,i3},k1,q1](component[H,{i3,i2},k2,q2]+I/2 component[G,{i3,i2},k2,q2])
        ),
        {i3,n},{k1,Abs[f[[i1]]-f[[i3]]],f[[i1]]+f[[i3]]},{q1,k1,-k1,-1},{k2,Abs[f[[i2]]-f[[i3]]],f[[i2]]+f[[i3]]},{q2,k2,-k2,-1}]+component[L,{i1,i2},k,q],
        {q,k,-k,-1}]}],
        {i1,n},{i2,n},{k,Abs[f[[i1]]-f[[i2]]],f[[i1]]+f[[i2]]}]
    ]

Options[InitialConditions]=DeleteCases[Options[DMVariables],TimeVariable->_]

InitialConditions[system:{_AtomicState..},InitialDM_?OperatorQ,t0_,opts___Rule]:=Module[{lhs,newopts,cev},
newopts=DeleteCases[Join[{opts},Options[LiouvilleEquation]],_->Automatic];
cev=ComplexExpandVariables/.newopts/.Options[DensityMatrix];
lhs=DensityMatrix[system,TimeVariable->t0,Sequence@@newopts];
If[cev||cev===Subscript,ExprToReIm[UpperTriangle[#],Sequence@@newopts]&,Flatten]@MapThread[Equal,{lhs,InitialDM},2]
]

InitialConditions[
	as:{_AtomicState..},
	es:{_Region..},
	InitialDM:{_?OperatorQ..},t0_,opts___Rule]:=
	Flatten[MapThread[InitialConditions[as,#2,t0,DMLabel->RegionLabel[#1],opts]&,{es,InitialDM}],1]

(*ExpandedInitialConditions[system:{_AtomicState..},InitialDM_?OperatorQ,t0_,opts___Rule]:=SeparateDensityMatrix[system,UpperTriangle@InitialConditions[system,InitialDM,t0,opts]]*)


End[]

EndPackage[]
