(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


Quiet[BeginPackage["AtomicDensityMatrix`WignerEckart`", { 
	"Units`", (* Units are loaded because BohrMagneton is used below. Think about updating to allow version 9 units functionality. *)
	"AtomicDensityMatrix`",
	"AtomicDensityMatrix`Common`",
	"AtomicDensityMatrix`CouplingCoefficients`",
	"AtomicDensityMatrix`AtomicSystem`"}],General::obspkg];
	
(* Exported symbols added here with SymbolName::usage *)  

(* ::Subsubsection:: *)
(*Wigner-Eckart theorem*)

(*Definition*)

WignerEckart::usage=""

(* ::Subsubsection:: *)
(*Reduced matrix elements*)

ReducedME::usage=""

OperatorOptions::usage=""

EvaluateHyperfineShift::usage=""

HyperfineShift::usage=""

Identity::usage=""

Zero::usage=""

GroundState::usage=""

ExcitedState::usage=""

Polarization::usage=""

StatisticalWeight::usage=""

Dipole::usage=""

ExpandDipoleRME::usage=""

SpontaneousEmission::usage=""

LandeGJ::usage=""

LandeGF::usage=""

DipoleExpansion::usage=""

MagneticMoment::usage=""

MagneticDipole::usage=""

Begin["`Private`"] (* Begin Private Context *) 

(* ::Subsection:: *)
(*Wigner-Eckart theorem*)

(*Returns matrix elements between Zeeman sublevels in terms of reduced matrix elements. Just returns ReducedME for a toy system.*)

(*Wigner-Eckart theorem [Var88, Sec. 13.1.1, Eq. (2)]:*)
(**)
(* \[LeftAngleBracket]\[Gamma] F M|\!*)
(*\*SubsuperscriptBox[\(T\), \(q\), \(\[Kappa]\)]|\[Gamma]' F' M'\[RightAngleBracket]=(-1)^(F-M) ({*)
(* {F, \[Kappa], F'},*)
(* {-M, q, M'}*)
(*})(\[Gamma] F||T^\[Kappa]||\[Gamma]' F')*)
(* or*)
(*  \[LeftAngleBracket]\[Gamma] F M|\!*)
(*\*SubsuperscriptBox[\(T\), \(q\), \(\[Kappa]\)]|\[Gamma]' F' M'\[RightAngleBracket]=(-1)^(2q) (F'M'\[Kappa] q|F M) (\[Gamma] F||T^\[Kappa]||\[Gamma]' F')/Sqrt[2F+1]*)

(*General case*)

(*wignerEckartFactor[s2_,{k_Integer,q_Integer},s1_] := wignerEckartFactor[s2,{k,q},s1] = 
	(2F[s2]+1)^(-1/2)CG[{F[s1],M[s1]},{k,q},{F[s2],M[s2]}]; 

WignerEckart[s2_AtomicState,{T_,k_Integer,q_Integer},s1_AtomicState]:=
    If[AtomicSystemType[s2] === "Toy",
        ReducedME[s2,{T,k},s1],
        wignerEckartFactor[s2,{k,q},s1] ReducedME[DeleteParameter[s2,M],{T,k},DeleteParameter[s1,M]] // Expand
    ]*)

wignerEckartFactor[{F1_,M1_},{k_,q_},{F2_,M2_}] := wignerEckartFactor[{F1,M1},{k,q},{F2,M2}] = 
	(2F2+1)^(-1/2)CG[{F1,M1},{k,q},{F2,M2}]; 

WignerEckart[s2_AtomicState,{T_,k_Integer,q_Integer},s1_AtomicState]:=
    If[AtomicSystemType[s2] === "Toy",
        ReducedME[s2,{T,k},s1],
        wignerEckartFactor[{F[s1],M[s1]},{k,q},{F[s2],M[s2]}] ReducedME[DeleteParameter[s2,M],{T,k},DeleteParameter[s1,M]] // Expand
    ]

(*The single component returned for a Toy system is wrapped in curly brackets so that code will work properly when Toy systems are changed to real systems.*)
(*For example, a square matrix inside one pair of curly brackets is a zero-dimensional Cartesian tensor, so ToCartesian returns the input. If we left off the brackets, then ToCartesian would choke on it.*)

WignerEckart[s2_AtomicState,{T_,k_Integer},s1_AtomicState]:=
    If[AtomicSystemType[s2] === "Toy",
        {WignerEckart[s2,{T,k,0},s1]},
        {WignerEckart[s2,{T,k,#},s1]& /@ Range[k,-k,-1]}
    ]

WignerEckart[sys:{_AtomicState..},{T_,k_Integer}]:=
    If[AtomicSystemType[sys] === "Toy",
        {WignerEckart[sys,{T,k,0}]},
        {WignerEckart[sys,{T,k,#}]& /@ Range[k,-k,-1]}
    ]

WignerEckart[sys:{_AtomicState..},Tn_] := DirectSelfProduct[WignerEckart[#1,Tn,#2]&,sys]

WignerEckart[sys1:{_AtomicState..},Tn_,sys2:{_AtomicState..}] := DirectProduct[WignerEckart[#1,Tn,#2]&,sys1,sys2]

WignerEckart[j_?HalfIntegerQ,Tn_] := WignerEckart[ZeemanSystem[j],Tn]

(*Reduced matrix elements*)

(*Polarization operator*)

ReducedME[s1_AtomicState,{Polarization,k_Integer},s2_AtomicState] := KroneckerDelta[StateLabel[s1],StateLabel[s2]] Sqrt[2k+1]

ReducedME[s1_AtomicState,{Polarization[{s1p_,f1_?NumericQ},{s2p_,f2_?NumericQ}],k_Integer},s2_AtomicState] :=
	KroneckerDelta@@Transpose@{{StateLabel[s1],s1p},{StateLabel[s2],s2p},{F[s1],f1},{F[s2],f2}} Sqrt[2k+1]

ReducedME[s1_AtomicState,{Polarization[s1p_,s2p_],k_Integer},s2_AtomicState]:=
	KroneckerDelta[StateLabel[s1],s1p]KroneckerDelta[StateLabel[s2],s2p] Sqrt[2k+1]

(*Hyperfine*)

(*Special case, statistical weight operator*)

ReducedME[s1_AtomicState,{StatisticalWeight,0},s2_AtomicState]/;AtomicSystemType[s1]==="HyperfineZeeman":=
	(-1)^(J[s1]+NuclearSpin[s1]+F[s2]) Sqrt[(2F[s1]+1)(2F[s2]+1)]SJS[{J[s1],F[s1],NuclearSpin[s1]},{F[s2],J[s2],0}] *
	(1/(2NuclearSpin[s1]+1) ReducedME[DeleteParameter[s1,F,NuclearSpin],{StatisticalWeight,0},DeleteParameter[s2,F,NuclearSpin]])

(*an angular coefficient*)

hyperfineK[J_,k_,F_]=(F+1)F-(k+1)k-(J+1)J

(*The energy of a given hyperfine level*)

EvaluateHyperfineShift[expr_] := expr/.HyperfineShift[J_,k_,F_,HyperfineA_,HyperfineB_]:> 
	If[J==1/2||k==1/2||J==0||k==0,
		1/2 HyperfineA hyperfineK[J,k,F],
		1/2 HyperfineA hyperfineK[J,k,F]+
			(3 HyperfineB*(hyperfineK[J,k,F] (hyperfineK[J,k,F]+1)-4/3 (k+1)k (J+1)J))/(8 (k (2 k-1) J*(2 J-1)))
	]

HyperfineShift[s_AtomicState] := EvaluateHyperfineShift@HyperfineShift[J[s],NuclearSpin[s],F[s],HyperfineA[s],HyperfineB[s]];

HyperfineShift[s:{_AtomicState..}] := HyperfineShift /@ s;

hyperfineRMEFactor[s1_,k_Integer,s2_] := hyperfineRMEFactor[s1,k,s2] = 	
	(-1)^(J[s1]+NuclearSpin[s1]+F[s2]+k) Sqrt[(2F[s1]+1)(2F[s2]+1)]SJS[{J[s1],F[s1],NuclearSpin[s1]},{F[s2],J[s2],k}]; 

ReducedME[s1_AtomicState,{Energy,0},s2_AtomicState] /; AtomicSystemType[s1]==="HyperfineZeeman":=
	With[{sJ = DeleteParameter[{s1,s2},F,NuclearSpin]},
			hyperfineRMEFactor[s1,0,s2] *
				(ReducedME[sJ[[1]],{Energy,0},sJ[[2]]] + HyperfineShift[s1] ReducedME[sJ[[1]],{Identity,0},sJ[[2]]])
	]
	
(* ReducedME for hyperfine structure, assuming the operator commutes with nuclear spin *)
(* (I J_1 F_1||T^k||I J_2 F_2) = (-1)^(J_1+I+F_2+k) Sqrt[(2 F_1+1)(2 F_2+1)] SJS[{J_1,F_1,I},{F_2,J_2,k}] (J_1||T^k||J_2) *)
(* Sobelman Eq. (4.175), p. 84 *)

ReducedME[s1_AtomicState,{T_,k_Integer},s2_AtomicState] /; AtomicSystemType[s1] === "HyperfineZeeman" :=  
	ReducedME[s1,{T,k},s2] =  
		hyperfineRMEFactor[s1,k,s2] ReducedME[DeleteParameter[s1,F,NuclearSpin],{T,k},DeleteParameter[s2,F,NuclearSpin]]

(*Zero*)

ReducedME[s1_AtomicState,{Zero,0},s2_AtomicState] = 0;

(*Identity operator*)

(*[Var88, Sec. 13.1.1, Eq. (3)]*)

ReducedME[s1_AtomicState,{Identity,0},s2_AtomicState] := KroneckerDelta[StateLabel[s1],StateLabel[s2]] Sqrt[2J[s1]+1]

(*Ground state operator*)

ReducedME[s1_AtomicState,{GroundState,0},s2_AtomicState] := Boole@GroundStateQ[s1] ReducedME[s1,{Identity,0},s2]

(*Excited state operator*)

ReducedME[s1_AtomicState,{ExcitedState,0},s2_AtomicState] := Boole@Not@GroundStateQ[s1] ReducedME[s1,{Identity,0},s2]

(*Natural width operator*)

ReducedME[s1_AtomicState,{NaturalWidth,0},s2_AtomicState] := NaturalWidth[s1] ReducedME[s1,{Identity,0},s2]

(*Statistical weight operator*)

ReducedME[s1_AtomicState,{StatisticalWeight,0},s2_AtomicState] := 1/(2J[s1]+1) ReducedME[s1,{Identity,0},s2]

(*Unperturbed energy operator*)

ReducedME[s1_AtomicState,{Energy,0},s2_AtomicState] := Energy[s1]ReducedME[s1,{Identity,0},s2]

(*E1 dipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

ReducedME[s1_AtomicState,{Dipole,1},s2_AtomicState] /; 
	Parity[s1]==Parity[s2] || StateLabel[s1]==StateLabel[s2] :=
		ReducedME[s1,{Dipole,1},s2] = 0;

(*Reorder states, using (\[Gamma] J||d||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||d||\[Gamma] J) *)
(*If states are in the right order, replace expression with a placeholder whose value can be set later*)

ReducedME[s1_AtomicState,{Dipole,1},s2_AtomicState] :=
	ReducedME[s1,{Dipole,1},s2] = 
		If[OrderedQ[{StateLabel[s1],StateLabel[s2]}],
			ReducedME[StateLabel[s1],{Dipole,1},StateLabel[s2]],
			If[AtomicSystemType[s1]==="Toy", 1, (-1)^(J[s1]-J[s2])] ReducedME[StateLabel[s2],{Dipole,1},StateLabel[s1]]
		]

DipoleExpansion[s1_AtomicState,s2_AtomicState]:=
	ReducedME[StateLabel[s1],{Dipole,1},StateLabel[s2]]->
		Which[
			BranchingRatio[s2][s1]===0&&BranchingRatio[s1][s2]=!=0,
				Sqrt[3]/2 Sqrt[((1+2 J[s2])NaturalWidth[s2]BranchingRatio[s1][s2])/(Energy[s2]-Energy[s1])^3],
			BranchingRatio[s1][s2]===0&&BranchingRatio[s2][s1]=!=0,
				Sqrt[3]/2 Sqrt[((1+2 J[s1])NaturalWidth[s1]BranchingRatio[s2][s1])/(Energy[s1]-Energy[s2])^3],
			True,
				ReducedME[StateLabel[s1],{Dipole,1},StateLabel[s2]]
		]

ExpandDipoleRME[syst_,expr_]:=
	expr/.Flatten@DirectSelfProduct[DipoleExpansion,Union[syst,SameTest->(StateLabel[#1]===StateLabel[#2]&)]]

(*Spontaneous emission*)

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState]/;StateLabel[s1]===StateLabel[s2]=0;

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState]/;Parity[s1]===Parity[s2]=0;

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState]/;!OrderedQ[{StateLabel[s1],StateLabel[s2]}]:=
	If[AtomicSystemType[s1]==="Toy", 1, (-1)^(J[s1]-J[s2])] ReducedME[s2,{SpontaneousEmission,1},s1]

ReducedME[s1_AtomicState,{SpontaneousEmission,1},s2_AtomicState]:=
	If[LevelPosition[s1]===Lower,
		Sqrt[(2J[s2]+1)NaturalWidth[s2]BranchingRatio[StateLabel[s1]][s2]],
		Sqrt[(2J[s1]+1)NaturalWidth[s1]BranchingRatio[StateLabel[s2]][s1]]
	]

(*Angular momentum operator*)

ReducedME[s1_AtomicState,{J,1},s2_AtomicState]:=
	Sqrt[J[s1](J[s1]+1)(2J[s1]+1)]KroneckerDelta[StateLabel[s1],StateLabel[s2]];

(*Magnetic moment operator*)

(*g factors*)

LandeGJ[0,L_,S_]=0;
LandeGJ[J_,L_,S_]=1+((J+1)J+(S+1)S-(L+1)L)/(2(J+1)J);
LandeGJ[s_AtomicState]:=LandeGJ[J[s],L[s],S[s]];
SetAttributes[LandeGJ,Listable];

LandeGF[0,J_,L_,S_,k_]=0;
LandeGF[F_,J_,L_,S_,k_]=LandeGJ[J,L,S] ((F+1)F+(J+1)J-(k+1)k)/(2(F+1)F);
LandeGF[s_AtomicState]:=LandeGF[F[s],J[s],L[s],S[s],NuclearSpin[s]];
SetAttributes[LandeGF,Listable];

(* The magnetic moment operator also has nonzero matrix elements for Delta J = 1, which are neglected here. 
Should take these into account, at least for resonant fields.*)

ReducedME[s1_AtomicState,{MagneticMoment,1},s2_AtomicState]:=
	-BohrMagneton Which[
		DefinedQ[s1,GFactor], GFactor[s1],
		DefinedQ[s1,{J,L,S}], LandeGJ[s1],
		True, GFactor[s1]
		] ReducedME[s1,{J,1},s2]

(* ::Subsubsection:: *)
(*MagneticDipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

(*ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState] /;
	Parity[s1]!=Parity[s2]||StateLabel[s1]==StateLabel[s2] := 0*)

(*Reorder states, using (s1 J||MagneticDipole||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||MagneticDipole||\[Gamma] J) *) 
(* NEED TO CHECK IF THIS IS CORRECT!*)

(*ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState] /; 
	!OrderedQ[{StateLabel[s1],StateLabel[s2]}] :=
		(-1)^(J[s1]-J[s2]) ReducedME[s2,{MagneticDipole,1},s1]*)

(* ::Subsubsection:: *)
(*Polarizability operator*)

(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[\!\(TraditionalForm\`*)
(*\*SubscriptBox[\(H\), \(Stark\)] = {\**)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"]\[CircleTimes]\**)
(*SuperscriptBox[*)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], "*"]}\), \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

alphaFactor[k_Integer,j_?HalfIntegerQ] /; k>=2j+1 := 0

alphaFactor[k_Integer,j_?HalfIntegerQ] := alphaFactor[k,j] = 
	Simplify[-(1/2)(Sqrt[2j+1])/(CG[{1,Mod[k,2]},{1,-Mod[k,2]},{k,0}]CG[{j,j},{k,0},{j,j}]), Element[2j,Integers]]

ReducedME[s1_AtomicState,{Polarizability,k_Integer},s2_AtomicState]:=
	alphaFactor[k,J[s1]]Polarizability[k][s1]KroneckerDelta[StateLabel[s1],StateLabel[s2]]

(* ::Subsubsection:: *)
(*TransitionPolarizability operator*)

ReducedME[s1_AtomicState,{TransitionPolarizability,k_Integer},s2_AtomicState]:=
	Switch[k,0,-Sqrt[3],1,Sqrt[2],2,1]TransitionPolarizability[k][s1,s2](1-KroneckerDelta[StateLabel[s1],StateLabel[s2]])

End[] (* End Private Context *)

EndPackage[]